class.TitleRunner(Ent) -- Pseudo-runner

local function random_audio(r) r.tape = tapes[math.random(#tapes)] end
local function random_generator(r) r.generator = generators[math.random(#generators)] end
local function return_to_title()
	local run = TitleRunner()
	next_runner(run)
end
local function register_return()
	engine.on_esc = return_to_title
end

artmode = 1

-- ART MODES: 1 Practice 2 Skill 3 Gallery
function TitleRunner:_init(spec)
	pull(self, {
		pressed={
			[engine.sdl.SDL_SCANCODE_P]=function(self) artmode = 1 end,
			[engine.sdl.SDL_SCANCODE_E]=function(self) artmode = 2 end,
			[engine.sdl.SDL_SCANCODE_G]=function(self) artmode = 3 end,
			[engine.sdl.SDL_SCANCODE_S]=function(self)
				local run = FreeRunner()
				random_audio(run) random_generator(run)
				table.insert(doom, function()
					register_return()
					next_runner(run)
				end)
			end,
			[engine.sdl.SDL_SCANCODE_1]=self.load,
			[engine.sdl.SDL_SCANCODE_2]=self.load,
			[engine.sdl.SDL_SCANCODE_3]=self.load,
			[engine.sdl.SDL_SCANCODE_4]=self.load,
			[engine.sdl.SDL_SCANCODE_5]=self.load,
			[engine.sdl.SDL_SCANCODE_6]=self.load,
			[engine.sdl.SDL_SCANCODE_7]=self.load,
			[engine.sdl.SDL_SCANCODE_8]=self.load,
			[engine.sdl.SDL_SCANCODE_9]=self.load,
		}
	})
	self:super(spec)
end

local lesson_titles={
	"Getting started: The blank page",
	"Still life drawing",
	"How to draw a landscape",
	"Introduction to portrait drawing",
	"Mastering facial expressions",
	"Learning from the greats",
}
local lesson_pngs = {
	"resource/stage/lesson_0.png",
	"resource/stage/lesson_1.png",
	"resource/stage/lesson_2.png",
	"resource/stage/lesson_3.png",
	"resource/stage/lesson_4.png",
	"resource/stage/lesson_5.png",
	"resource/stage/scribble.png",
	"resource/stage/pegasister.png",
}

function TitleRunner:insert(prio)
	Ent.insert(self, prio)
	engine.clear_screen()
end

function TitleRunner:load(what)
	local idx = what - engine.sdl.SDL_SCANCODE_1 + 1
	local found = lesson_pngs[idx]
	local title = lesson_titles[idx]
	if found then
		local run = CoolRunner()
		random_audio(run)
		random_generator(run)
		run.target = found
		
		if artmode == 2 then run.limit = 10*1000 end
		
		if not title then
			title = ""
			for i=1,16 do title = title .. (math.random()<0.1 and " " or string.char(math.random(127))) end
		end
		run.title = string.format("Lesson %d: %s", idx, title)
		
		table.insert(doom, function()
			register_return()
			next_runner(run)
		end)
	end
end

function TitleRunner:onTick()
	local modeline = "Pick mode:"

	local titles = {
		"BECOME A GREAT ARTIST IN JUST 10 SECONDS",
		"With Andi and Michael",
		"",
		"Instructions: Press keys on key board.",
		"Remember DELETE to rewind",
		"",
		"Pick:       (S)ketchpad      (T)utorial",
		"",
		"--------------Art lessons--------------",
		string.format("Pick mode: (P)%s  (E)%s (G)%s",
			artmode==1 and "RACTICE" or "ractice",
			artmode==2 and "XAM" or "xam",
			artmode==3 and "ALLERY" or "allery"),
		"",
		"Pick lesson:",
		""
	}
	
	for i,v in ipairs(lesson_titles) do
		table.insert(titles, string.format("(%d) %s", i, v))
	end
	table.insert(titles,"")
	
	local modes = {"In PRACTICE mode there is no time limit",
	               "Put it to the test in a ten-second EXAM",
   				   "GALLERY of your Exam mode masterpieces!"}
	table.insert(titles, modes[artmode])
	
	local longest = #modes[2]
--	for i,v in ipairs(modes) do if #v > longest then longest = #v end end
	local blockwidth = engine.fw*longest
	local blockheight = engine.fh*#titles
	
	local bx,by = (engine.sw-blockwidth)/2, (engine.sh-blockheight)/2
--	print({blockwidth,blockheight,engine.sw,engine.sh,bx,by})
	engine.draw_text( bx, by, titles )
end